#include <windows.h>
#include <WinIoCtl.h>
#include <stdio.h>
#include "args.h"

#ifdef _DEBUG
#define KdPrint(x)  {printf x;}
#else
#define KdPrint(x)  {;}
#endif

typedef struct _TBB_MSG_HDR {
  UCHAR Signature[4]; // 21 09 70 19 30 00 00 00
  ULONG HdrLength; // 0x00000030
  UCHAR hz1[8];
  ULONG hz2;
  ULONG Flags;
  ULONG hz3[3];
  ULONG MsgLength; // without header
  ULONG hz4;
  ULONG tail0;
} TBB_MSG_HDR, *PTBB_MSG_HDR;

#define MSG_FLAG_DEL   0x01
#define MSG_FLAG_READ  0x02
#define MSG_FLAG_ANSWERED 0x04
#define MSG_FLAG_PARKED   0x08
#define MSG_FLAG_FLAGGED  0x40

#define TB_HDR_SZ  sizeof(TBB_MSG_HDR)
#define MAIL_HDR_SZ  sizeof("Return-Path:")

ULONG iobs = 64*1024;
ULONG bs = 512;

__int64 m_ptr;
__int64 next_m_ptr;
ULONG m_count;
ULONG dm_count;
ULONG prev_msg_len = 0;
ULONG skip_msg=0;
ULONG lim_msg=0;

__int64 copy_count = 0;
__int64 mlt = 0;

BOOLEAN no_cache = 0;
BOOLEAN dump_mode = 0;
BOOLEAN unix_mode = 1;
BOOLEAN skip_deleted = 1;
BOOLEAN o_writing = 1;
BOOLEAN list_msg = 0;

PWCHAR in_nameW = NULL;
PWCHAR out_nameW = NULL;
PWCHAR log_nameW = NULL;
PWCHAR fill_stringW = NULL;

PCHAR in_nameA = NULL;
PCHAR out_nameA = NULL;
PCHAR log_nameA = NULL;
PCHAR fill_stringA = NULL;


char* buff_template = NULL;

int new_target = 0;

CHAR LogBuffer[2048];
CHAR MailRowBuffer[2048];

void print_help() {
    printf("Usage:\n"
           "  tbbcopy.exe [-<switches>] <source file> <target file>\n"
           "      or\n"
           "  tbbcopy.exe <dd-style options>\n"
           "Switches:\n"
           "  -o[b|k|m|g] <offset>  offset in bytes (or optionally in blocks, KBytes, MBytes or GBytes)\n"
           "                      to start recovery from\n"
/*           "  -c[b|k|m|g] <count>   number of bytes (or optionally in blocks, KBytes, MBytes or GBytes)\n"
           "                      to recover\n"*/
           "  -L <logfile>        log errors to <logfile>\n"
/*           "  -n                  use NO CACHE option for file/device I/O\n"
           "                        This option is automatically enabled for block devices\n"*/
           "  -l                  just dump e-mail list\n"
/*           "  -u                  extract to unix-mailbox\n"*/
           "  -d[+|-]             include (+) or exclude(-) deleted messages\n"
           "  -m <num>            start from message <num>\n"
           "  -M <num>            process not more than <num>\n"
           "  -?                  display this help\n"
           "dd-style options:\n"
           "  if=<source file>\n"
           "  of=<target file>\n"
           "  seek=<offset>[units]    default unit 'b'\n"
           "                            set SRC positions\n"
           "Units:\n"
           "  o - bytes (octets)      b - blocks\n"
           "  k - KBytes)             M - MBytes             G - GBytes\n"
           "Note:\n"
           "  1 KByte = 1024 bytes\n"
           "  1 MByte = 1024 KBytes = 1024*1024 bytes\n"
           "  if byte offset or byte count is specified, it will be rounded to blocksize\n"
           "Examples:\n"
           "  tbbcopy.exe -o 10M X:\\mail\\messages.tbb E:\\Temp\\Restored.tbb\n"
           "    will extract messages starting from byte offset 10Mb to end of file\n"
          );
    exit(0);
} // end print_help()

__int64
get_file_size(
    HANDLE h,
    BOOLEAN* is_dev
    )
{
    LONG hsz = 0;
    LONG lsz;
    __int64 sz;
    ULONG status;
    ULONG returned;
    DISK_GEOMETRY  DiskGeometry;

    if(is_dev) {
        *is_dev = FALSE;
    }

    lsz = SetFilePointer(h, 0, &hsz, FILE_END);
    if(lsz != -1 && hsz != -1) {
        ((PULONG)(&sz))[1] = hsz;
        ((PULONG)(&sz))[0] = lsz;
        return sz;
    }

    status = DeviceIoControl(h,
                             IOCTL_DISK_GET_DRIVE_GEOMETRY,
                             &DiskGeometry,sizeof(DISK_GEOMETRY),
                             &DiskGeometry,sizeof(DISK_GEOMETRY),
                             &returned,
                             FALSE);
    if(!status) {
        status = GetLastError();
        return -1;
    }

    if(is_dev) {
        *is_dev = TRUE;
    }
    sz = (DiskGeometry.Cylinders.QuadPart *
          DiskGeometry.TracksPerCylinder *
          DiskGeometry.SectorsPerTrack)*
          DiskGeometry.BytesPerSector;
    if(sz > 0) {
        bs = DiskGeometry.BytesPerSector;
        return sz;
    }

    return -1;
} // end get_file_size()

__int64
set_file_pointer(
    HANDLE h,
    __int64 sz
    )
{
    ULONG hsz = ((PULONG)(&sz))[1];
    ULONG lsz = ((PULONG)(&sz))[0];

    lsz = SetFilePointer(h, lsz, (PLONG)&hsz, FILE_BEGIN);
    if(lsz != -1 && hsz != -1) {
        ((PULONG)(&sz))[1] = hsz;
        ((PULONG)(&sz))[0] = lsz;
        return sz;
    }
    return 0;
} // end set_file_pointer()

int
set_file_size(
    HANDLE h,
    __int64 sz
    )
{
    set_file_pointer(h, sz);
    return SetEndOfFile(h);
} // end set_file_size()

int
write_output(
    HANDLE out_h,
    char* o_buff,
    ULONG* o_buff_ptr,
    char* data,
    ULONG xbs,
    __int64 *_o_ptr,
    BOOLEAN flush
    )
{
    ULONG rest;
    ULONG read_bytes;

    if(list_msg) {
        return 1;
    }

    if(!o_writing) {
      if(!flush) {
        return 0;
      }
      xbs = 0;
    }

    rest = iobs-(*o_buff_ptr);
    if(rest >= xbs) {
      memcpy(o_buff+(*o_buff_ptr), data, xbs);
      (*o_buff_ptr) += xbs;
    } else {
      while(xbs) {
        read_bytes = min(rest,xbs);
        write_output(out_h, o_buff, o_buff_ptr, data, read_bytes, _o_ptr, flush);
        xbs -= read_bytes;
        rest = iobs;
      }
      return 1;
    }
    if(!flush && ((*o_buff_ptr) & (iobs-1))) {
      return xbs;
    }

    rest = flush ? (*o_buff_ptr) : iobs;

    set_file_pointer(out_h, *_o_ptr);
    WriteFile(out_h, o_buff, rest, &read_bytes, NULL);
    if(read_bytes != rest) {
      read_bytes = GetLastError();
      printf("Can't write block at offset %I64d error %d\n", *_o_ptr, read_bytes);
      return -1;
    }
    (*o_buff_ptr) = 0;
    (*_o_ptr) += rest;
    return xbs;

} // end write_output()

PCHAR m_buff=NULL;
ULONG m_buff_sz=0;
ULONG m_buff_ptr=0;

int
store_msg(
    char* data,
    ULONG len,
    BOOLEAN flush
    )
{
    PCHAR m_buff_tmp;

    if(!m_buff) {
        m_buff = (char*)GlobalAlloc(GMEM_FIXED, iobs);
        if(!m_buff) {
            return 0;
        }
        m_buff_sz = iobs;
    }

    if(list_msg) {
        return 1;
    }

    if(m_buff_ptr+len > m_buff_sz) {
        m_buff_tmp = (char*)GlobalAlloc(GMEM_FIXED, m_buff_sz*3/2+len);
        if(!m_buff_tmp) {
            return 0;
        }
        memcpy(m_buff_tmp, m_buff, m_buff_ptr);
        GlobalFree(m_buff);
        m_buff = m_buff_tmp;
        m_buff_sz = m_buff_sz*3/2+len;
    }
    memcpy(m_buff+m_buff_ptr, data, len);
    m_buff_ptr += len;

    return len;

} // end store_msg()

void
fill_buffer(
    char* buff,
    ULONG bs,
    PWCHAR fill_stringW,
    PCHAR fill_stringA
    )
{
/*
    if(!fill_stringA && !fill_stringW) {
        memset(buff, 0, bs);
    } else {
        if(!buff_template) {
            int tail = bs;
            int d;
            char* p_buff = buff;

            buff_template = (char*)GlobalAlloc(GMEM_FIXED, iobs);
            if(!buff_template) {
                memset(buff, 0, bs);
                return;
            }

            p_buff = buff_template;
            tail = iobs;

            while(tail > 0) {
                if(GetVersion() & 0x80000000) {
                    d = _snprintf(p_buff, tail, "%s", fill_stringA);
                } else {
                    d = _snprintf(p_buff, tail, "%S", fill_stringW);
                }
                tail -= d;
                p_buff += d;
            }
        }
        memcpy(buff, buff_template, bs);
    }
*/
    memcpy(buff, buff+iobs, iobs);
    memset(buff+iobs, 0, iobs);
} // end fill_buffer()

#include "copy_fragment.h"

#pragma warning(push)               
#pragma warning(disable:4005)
#define XCHAR    CHAR
#define get_num  get_numA
#include "get_num.cpp.h"

#define XCHAR    WCHAR
#define get_num  get_numW
#include "get_num.cpp.h"
#pragma warning(pop)

void print_result() {
    printf("Processed %I64d bytes (%I64d blocks)\n"
           " %d messages found\n"
           " %d messages extracted ", copy_count, copy_count/bs, dm_count, m_count);
} // end print_result()

void main(void) {
    int i, j/*, n*/;
    WCHAR a;
    ULONG r;
    ULONG xbs;
    ULONG err = 0;
   
    int break_sw;
    int block_seek = 0;
    int block_offs = 0;
    int block_counter = 0;
    int block_iobs = 0;
    __int64 seek = 0;
    //__int64 offs = 0;
    __int64 counter = 0;
    __int64 f_size;
    __int64 f_ptr;
    __int64 o_ptr;
    __int64 out_size;
    
    ULONG max_cached; // Gb

    in_nameW = NULL;    
    out_nameW = NULL;   
    log_nameW = NULL;   
    fill_stringW = NULL;

    in_nameA = NULL;    
    out_nameA = NULL;   
    log_nameA = NULL;   
    fill_stringA = NULL;
    
    bs = 2048;     
    iobs = 64*1024;

    char* buff;
    char* o_buff;
    ULONG o_buff_ptr;
    ULONG read_bytes;
    int ok;
    __int64 p;
    int prc, prc0;

    WCHAR* CmdLineW;
    PWCHAR* argvW;
    CHAR*  CmdLineA;
    PCHAR* argvA;
    int argc;

    HANDLE in_h;
    HANDLE out_h;
    HANDLE log_h = NULL;

    BOOLEAN in_is_dev = FALSE;
    BOOLEAN out_is_dev = FALSE;
    BOOLEAN out_is_nul = FALSE;

    printf("The Bat TBB recovery utility v1.0 (c) by Alter, 2012\n"
           "Home site http://alter.org.ua\n");

    if(GetVersion() & 0x80000000) {
        printf("Warning: running under Windows 9x family\n"
               "  You may experience 2Gb or 4Gb limitation.\n");
//        exit(1);
        max_cached = 2;

        CmdLineA = GetCommandLineA();
        argvA = CommandLineToArgvA(CmdLineA, &argc);

#pragma warning(push)               
#pragma warning(disable:4005)
#define argv     argvA
#define in_name  in_nameA
#define out_name out_nameA
#define log_name log_nameA
#define in_devname   in_devnameA
#define out_devname  out_devnameA
#define fill_string  fill_stringA
#define xsprintf sprintf
#define xcslen   strlen
#define xscanf   sscanf
#define xprintf  printf
#define xstrcmp  strcmp
#define xstrncmp  strncmp
#define CreateFileX   CreateFileA
#define XT(a)    a
#define XCHAR    CHAR
#define get_num  get_numA
#define open_all open_allA
#pragma warning(pop)

        #include "parse_args.cpp.h"

        #include "open_all.cpp.h"

    } else {
        CmdLineW = GetCommandLineW();
        argvW = CommandLineToArgvW(CmdLineW, &argc);

        max_cached = 100;

#pragma warning(push)               
#pragma warning(disable:4005)
//#pragma warning(disable:6311)
#define argv     argvW
#define in_name  in_nameW
#define out_name out_nameW
#define log_name log_nameW
#define in_devname   in_devnameW
#define out_devname  out_devnameW
#define fill_string  fill_stringW
#define xsprintf swprintf
#define xcslen   wcslen
#define xscanf   swscanf
#define xprintf  wprintf
#define xstrcmp  wcscmp
#define xstrncmp  wcsncmp
#define CreateFileX   CreateFileW
#define XT(a)    L##a
#define XCHAR    WCHAR
#define get_num  get_numW
#define open_all open_allW
#pragma warning(pop)

        #include "parse_args.cpp.h"

        #include "open_all.cpp.h"

    }

    buff = (char*)GlobalAlloc(GMEM_FIXED, iobs*2);
    if(!buff) {
        printf("Can't alloc %d Kbytes for temporary buffer\n"
               "exiting...\n", iobs/512);
        exit (-2);
    }
    o_buff = (char*)GlobalAlloc(GMEM_FIXED, iobs);
    o_buff_ptr = 0;

    f_size = get_file_size(in_h, &in_is_dev);
    if(!f_size) {
        if(GetVersion() & 0x80000000) {
            printf("Empty source file %s\n"
                   "exiting...\n", out_nameA);
        } else {
            wprintf(L"Empty source file %s\n"
                   L"exiting...\n", out_name);
        }
        exit (-4);
    }
    if(f_size < 0) {
        if(GetVersion() & 0x80000000) {
            printf("Can't get source file size: %s\n"
                   "exiting...\n", out_nameA);
        } else {
            wprintf(L"Can't get source file size: %s\n"
                   L"exiting...\n", out_name);
        }
        exit (-4);
    }

    // block devices and too long files are not necessary to be cached
    if(!no_cache &&
       (in_is_dev || out_is_dev || (f_size >= (__int64)((unsigned __int64)1024*1024*1024*(unsigned __int64)max_cached))) ) {
        // reopen with NO_CACHE

        if(in_is_dev || out_is_dev) {
            printf("Force using NO_CACHE on block devices\n");
        } else {
            printf("Force using NO_CACHE on large file\n");
        }

        no_cache = 1;

        CloseHandle(in_h);
        in_h = NULL;
        if(out_h) {
            CloseHandle(out_h);
            out_h = NULL;
        }

        if(GetVersion() & 0x80000000) {
#pragma warning(push)               
#pragma warning(disable:4005)
#define open_all open_allA
#pragma warning(pop)
        #include "open_all.cpp.h"
        } else {
#pragma warning(push)               
#pragma warning(disable:4005)
#define open_all open_allW
        #include "open_all.cpp.h"
#pragma warning(pop)
        }
    }

    out_size = get_file_size(out_h, &out_is_dev);

    if(in_is_dev || out_is_dev) {
        if(GetVersion() & 0x80000000) {
            printf("Block device not supported. \n"
                   "exiting...\n");
        } else {
            wprintf(L"Block device %s too small: %I64d bytes, required %I64d\n"
                    L"exiting...\n");
        }
        exit (-5);
    }

    seek = seek & ~((unsigned __int64)(bs-1));
    counter = (counter + bs - 1) & ~((unsigned __int64)(bs-1));

    f_ptr = seek;
    o_ptr = 0;
    m_ptr = 0;
    next_m_ptr = 0;
    m_count = 0;
    dm_count = 0;

    if(GetVersion() & 0x80000000) {
        printf("%s -> %s\n%I64d bytes\n",
               in_nameA, out_nameA, f_size);
    } else {
        wprintf(L"%s -> %s\n%I64d bytes\n",
               in_name, out_name, f_size);
    }
    printf("Starting from offset %I64d bytes\n",
           seek);
/*    if(counter) {
        printf("Recovery range %I64d bytes (%I64d blocks)\n",
               counter, counter/bs);
    }
*/
    if(counter) {
        p = copy_count/50;
    } else {
        p = f_size/50;
    }
    if(!p) p = 1;
    prc0 = -1;

    while(f_ptr <= f_size-bs) {
        if(counter && copy_count >= counter) {
            print_result();
            exit(0);
        }
        if(counter) {
            prc = (int)(copy_count/p);
        } else {
            prc = (int)(f_ptr/p);
        }
        if( prc0 != prc) {
            prc0 = prc;
            printf(".");
        }
        if(counter && (counter-copy_count < iobs)) {
            xbs = (ULONG)(counter-copy_count);
        } else
        if(f_size - f_ptr < iobs) {
            xbs = bs;
        } else {
            xbs = iobs;
        }
        ok = fetch_fragment(in_h, out_h, log_h, buff, o_buff, &o_buff_ptr, xbs, &r, f_ptr, &o_ptr);

        if(ok <= 0) {
            printf("Stop on output error\n");
            print_result();
            exit(-9);
        }
        f_ptr += xbs;
        //o_ptr += xbs;
        copy_count += xbs;

        if(lim_msg && m_count>lim_msg) {
          f_size = f_ptr;
          break;
        }
    }

    printf("\n");
    if(f_size < f_ptr ||
        f_size - f_ptr > bs) {
        printf("Can't copy file tail (%I64d bytes)\n", f_size - f_ptr);
        exit(-7);
    }
    xbs = (ULONG)(min(iobs, f_size - f_ptr));
    if(!xbs || (counter && copy_count >= counter)) {
        print_result();
        exit(0);
    }

    if((xbs & (bs-1)) && no_cache && !(in_is_dev || out_is_dev)) {
        CloseHandle(in_h);
        in_h = NULL;
        if(out_h) {
            CloseHandle(out_h);
            out_h = NULL;
        }

        if(GetVersion() & 0x80000000) {
#pragma warning(push)               
#pragma warning(disable:4005)
#define open_all open_allA
#pragma warning(pop)
        #include "open_all.cpp.h"
        } else {
#pragma warning(push)               
#pragma warning(disable:4005)
#define open_all open_allW
        #include "open_all.cpp.h"
#pragma warning(pop)
        }
    }

    if(counter && (counter-copy_count < iobs)) {
        xbs = (ULONG)(counter-copy_count);
    } else
    if(f_size - f_ptr < iobs) {
        xbs = (ULONG)(f_size - f_ptr);
        if(in_is_dev || out_is_dev) {
          xbs &= ~(bs-1);
        }
    } else {
        xbs = iobs;
    }
    ok = fetch_fragment(in_h, out_h, log_h, buff, o_buff, &o_buff_ptr, xbs, &r, f_ptr, &o_ptr);
    if(ok <= 0) {
        printf("Stop on output error\n");
        print_result();
        exit(-9);
    }
    //#define set_filler set_filler2
    //#include "fetch_fragment.h"
    f_ptr += xbs;
    //o_ptr += xbs;
    copy_count += xbs;

    print_result();
    exit(0);
} // end main()
	